@@ -1,3 +1,20 @@
+# Licensed to the Apache Software Foundation (ASF) under one
+# or more contributor license agreements.  See the NOTICE file
+# distributed with this work for additional information
+# regarding copyright ownership.  The ASF licenses this file
+# to you under the Apache License, Version 2.0 (the
+# "License"); you may not use this file except in compliance
+# with the License.  You may obtain a copy of the License at
+#
+#   http://www.apache.org/licenses/LICENSE-2.0
+#
+# Unless required by applicable law or agreed to in writing,
+# software distributed under the License is distributed on an
+# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
+# KIND, either express or implied.  See the License for the
+# specific language governing permissions and limitations
+# under the License.
+
 ## TODO
 ## this class is here because of https://github.com/gfx/p5-Mouse/pull/67
 ## once 2.4.7 version of Mouse in Ubuntu for affected Perl version
@@ -6,7 +23,7 @@
 package AI::MXNet::Module::Private;
 use Mouse;
 has [qw/_param_names _fixed_param_names
-        _aux_names _data_names _label_names
+        _aux_names _data_names _label_names _state_names
         _output_names _arg_params _aux_params
         _params_dirty _optimizer _kvstore
          _update_on_kvstore _updater _work_load_list
@@ -15,25 +32,28 @@ has [qw/_param_names _fixed_param_names
 ] => (is => 'rw', init_arg => undef);
 
 package AI::MXNet::Module;
+use AI::MXNet::NS;
 use AI::MXNet::Base;
 use AI::MXNet::Function::Parameters;
 use List::Util qw(max);
+use Data::Dumper ();
 use Mouse;
 
-=head2 _create_kvstore
-
-    Create kvstore
-    This function select and create a proper kvstore if given the kvstore type
-
-    Parameters
-    ----------
-    kvstore : KVStore or str
-        The kvstore
-    num_device : int
-        The number of devices
-    arg_params : dict of str to NDArray
-        Model parameter, dict of name to NDArray of net's weights.
-=cut
+func _create_sparse_kvstore(Maybe[Str|AI::MXNet::KVStore] $kvstore)
+{
+    # always update on kvstore
+    my $update_on_kvstore = 1;
+    my $kv;
+    if(blessed $kvstore)
+    {
+        $kv = $kvstore;
+    }
+    else
+    {
+        $kv = AI::MXNet::KVStore->create($kvstore);
+    }
+    return ($kv, $update_on_kvstore);
+}
 
 func _create_kvstore(
     Maybe[Str|AI::MXNet::KVStore] $kvstore,
@@ -76,38 +96,30 @@ func _create_kvstore(
     return ($kv, $update_on_kvstore);
 }
 
-=head2 _initialize_kvstore
-
-    Initialize kvstore
-=cut
-
 func _initialize_kvstore(
     AI::MXNet::KVStore           :$kvstore,
-    ArrayRef[AI::MXNet::NDArray] :$param_arrays,
     HashRef[AI::MXNet::NDArray]  :$arg_params,
     ArrayRef[Str]                :$param_names,
-    Bool                         :$update_on_kvstore
+    Bool                         :$update_on_kvstore,
+    ArrayRef[AI::MXNet::NDArray]|ArrayRef[ArrayRef[AI::MXNet::NDArray]] :$param_arrays
 )
 {
     enumerate(sub{
         my ($idx, $param_on_devs) = @_;
-        $kvstore->init($idx, $arg_params->{ $param_names->[$idx] });
+        my $name = $param_names->[$idx];
+        $kvstore->init($name, $arg_params->{ $name });
         if($update_on_kvstore)
         {
-            $kvstore->pull($idx, $param_on_devs, priority => -$idx);
+            $kvstore->pull($name, out => $param_on_devs, priority => -$idx);
         }
     }, $param_arrays);
 }
 
-=head2 _update_params_on_kvstore
-
-    Perform update of param_arrays from grad_arrays on kvstore.
-=cut
-
-method _update_params_on_kvstore(
-    ArrayRef[AI::MXNet::NDArray] $param_arrays,
-    ArrayRef[AI::MXNet::NDArray] $grad_arrays,
-    AI::MXNet::KVStore           $kvstore
+func _update_params_on_kvstore(
+    ArrayRef[AI::MXNet::NDArray]|ArrayRef[ArrayRef[AI::MXNet::NDArray]] $param_arrays,
+    ArrayRef[AI::MXNet::NDArray]|ArrayRef[ArrayRef[AI::MXNet::NDArray]] $grad_arrays,
+    AI::MXNet::KVStore           $kvstore,
+    ArrayRef[Str]                $param_names
 )
 {
     enumerate(sub{
@@ -116,23 +128,21 @@ method _update_params_on_kvstore(
         {
             return;
         }
+        my $name = $param_names->[$index];
         # push gradient, priority is negative index
-        $kvstore->push($index, $grad_list, priority => -$index);
+        $kvstore->push($name, $grad_list, priority => -$index);
         # pull back the weights
-        $kvstore->pull($index, $arg_list, priority  => -$index);
+        $kvstore->pull($name, out => $arg_list, priority  => -$index);
     }, $param_arrays, $grad_arrays);
 }
 
-=head _update_params
-
-    Perform update of param_arrays from grad_arrays not on kvstore.
-=cut
 func _update_params(
     ArrayRef[ArrayRef[AI::MXNet::NDArray]] $param_arrays,
     ArrayRef[ArrayRef[AI::MXNet::NDArray]] $grad_arrays,
     AI::MXNet::Updater                     $updater,
     Int                                    $num_device,
-    Maybe[AI::MXNet::KVStore]              $kvstore=
+    Maybe[AI::MXNet::KVStore]              $kvstore=,
+    Maybe[ArrayRef[Str]]                   $param_names=
 )
 {
     enumerate(sub{
@@ -143,48 +153,23 @@ func _update_params(
         }
         if($kvstore)
         {
+            my $name = $param_names->[$index];
             # push gradient, priority is negative index
-            $kvstore->push($index, $grad_list, priority => -$index);
+            $kvstore->push($name, $grad_list, priority => -$index);
             # pull back the sum gradients, to the same locations.
-            $kvstore->pull($index, $grad_list, priority => -$index);
+            $kvstore->pull($name, out => $grad_list, priority => -$index);
         }
         enumerate(sub {
             my ($k, $w, $g) = @_;
             # faked an index here, to make optimizer create diff
             # state for the same index but on diff devs, TODO(mli)
-            # use a better solution latter
-            &{$updater}($index*$num_device+$k, $g, $w);
+            # use a better solution later
+            $updater->($index*$num_device+$k, $g, $w);
         }, $arg_list, $grad_list);
     }, $param_arrays, $grad_arrays);
 }
 
-=head2 load_checkpoint
-
-    Load model checkpoint from file.
-
-    Parameters
-    ----------
-    prefix : str
-        Prefix of model name.
-    epoch : int
-        Epoch number of model we would like to load.
-
-    Returns
-    -------
-    symbol : Symbol
-        The symbol configuration of computation network.
-    arg_params : dict of str to NDArray
-        Model parameter, dict of name to NDArray of net's weights.
-    aux_params : dict of str to NDArray
-        Model parameter, dict of name to NDArray of net's auxiliary states.
-
-    Notes
-    -----
-    - symbol will be loaded from ``prefix-symbol.json``.
-    - parameters will be loaded from ``prefix-epoch.params``.
-=cut
-
-func load_checkpoint(Str $prefix, Int $epoch)
+method load_checkpoint(Str $prefix, Int $epoch)
 {
     my $symbol = AI::MXNet::Symbol->load("$prefix-symbol.json");
     my %save_dict = %{ AI::MXNet::NDArray->load(sprintf('%s-%04d.params', $prefix, $epoch)) };
@@ -205,40 +190,24 @@ func load_checkpoint(Str $prefix, Int $epoch)
     return ($symbol, \%arg_params, \%aux_params);
 }
 
-=head2 new
+=head1 NAME
 
-    Module is a basic module that wrap a `Symbol`. It is functionally the same
-    as the `FeedForward` model, except under the module API.
-
-    Parameters
-    ----------
-    symbol : Symbol
-    data_names : list of str
-        Default is `('data')` for a typical model used in image classification.
-    label_names : list of str
-        Default is `('softmax_label')` for a typical model used in image
-        classification.
-    logger : Logger
-        Default is `logging`.
-    context : Context or list of Context
-        Default is `cpu()`.
-    work_load_list : list of number
-        Default `None`, indicating uniform workload.
-    fixed_param_names: list of str
-        Default `None`, indicating no network parameters are fixed.
+    AI::MXNet::Module - FeedForward interface of MXNet.
+    See AI::MXNet::Module::Base for the details.
 =cut
 
 extends 'AI::MXNet::Module::Base';
 
 has '_symbol'           => (is => 'ro', init_arg => 'symbol', isa => 'AI::MXNet::Symbol', required => 1);
-has 'data_names'        => (is => 'rw', isa => 'ArrayRef[Str]', default => sub { ['data'] });
-has 'label_names'       => (is => 'ro', isa => 'Maybe[ArrayRef[Str]]', default => sub { ['softmax_label'] });
-has 'work_load_list'    => (is => 'rw', isa => 'ArrayRef[Int]');
-has 'fixed_param_names' => (is => 'rw', isa => 'ArrayRef[Str]');
+has '_data_names'       => (is => 'ro', init_arg => 'data_names', isa => 'ArrayRef[Str]');
+has '_label_names'      => (is => 'ro', init_arg => 'label_names', isa => 'Maybe[ArrayRef[Str]]');
+has 'work_load_list'    => (is => 'rw', isa => 'Maybe[ArrayRef[Int]]');
+has 'fixed_param_names' => (is => 'rw', isa => 'Maybe[ArrayRef[Str]]');
+has 'state_names'       => (is => 'rw', isa => 'Maybe[ArrayRef[Str]]');
 has 'logger'            => (is => 'ro', default => sub { AI::MXNet::Logging->get_logger });
 has '_p'                => (is => 'rw', init_arg => undef);
 has 'context'           => (
-    is => 'ro', 
+    is => 'ro',
     isa => 'AI::MXNet::Context|ArrayRef[AI::MXNet::Context]',
     default => sub { AI::MXNet::Context->cpu }
 );
@@ -246,7 +215,11 @@ has 'context'           => (
 around BUILDARGS => sub {
     my $orig  = shift;
     my $class = shift;
-    return $class->$orig(symbol => $_[0]) if @_ == 1;
+    if(@_%2)
+    {
+        my $symbol = shift;
+        return $class->$orig(symbol => $symbol, @_);
+    }
     return $class->$orig(@_);
 };
 
@@ -267,53 +240,58 @@ sub BUILD
     }
     assert(@{ $work_load_list } == @{ $self->_p->_context });
     $self->_p->_work_load_list($work_load_list);
-    my @data_names  = @{ $self->data_names };
-    my @label_names = @{ $self->label_names//[] };
+    my @data_names  = @{ $self->_data_names//['data'] };
+    my @label_names = @{ $self->_label_names//['softmax_label'] };
+    my @state_names = @{ $self->state_names//[] };
     my $arg_names   = $self->_symbol->list_arguments;
-    my @input_names = (@data_names, @label_names);
+    my @input_names = (@data_names, @label_names, @state_names);
     my %input_names = map { $_ => 1 } @input_names;
     $self->_p->_param_names([grep { not exists $input_names{$_} } @{ $arg_names }]);
-    $self->_p->_fixed_param_names($self->fixed_param_names);
+    $self->_p->_fixed_param_names($self->fixed_param_names//[]);
+    $self->_p->_state_names(\@state_names);
     $self->_p->_aux_names($self->_symbol->list_auxiliary_states);
     $self->_p->_data_names(\@data_names);
     $self->_p->_label_names(\@label_names);
     $self->_p->_output_names($self->_symbol->list_outputs);
     $self->_p->_params_dirty(0);
-    $self->data_names($self->_p->_data_names);
+    $self->_check_input_names($self->_symbol, $self->_p->_data_names, "data", 1);
+    $self->_check_input_names($self->_symbol, $self->_p->_label_names, "label", 0);
+    $self->_check_input_names($self->_symbol, $self->_p->_state_names, "state", 1);
+    $self->_check_input_names($self->_symbol, $self->_p->_fixed_param_names, "fixed_param", 1);
 }
 
-sub Module { my $class = shift; return $class->new(@_) }
-sub BucketingModule { shift; return AI::MXNet::Module::Bucketing->new(@_) }
+method Module(@args) { return @args ?  __PACKAGE__->new(@args) : __PACKAGE__ }
+method BucketingModule(@args) { return AI::MXNet::Module::Bucketing->new(@args) }
 
-=head load
+=head2 load
 
         Create a model from previously saved checkpoint.
 
         Parameters
         ----------
-        prefix : str
+        $prefix : Str
             path prefix of saved model files. You should have
             "prefix-symbol.json", "prefix-xxxx.params", and
             optionally "prefix-xxxx.states", where xxxx is the
             epoch number.
-        epoch : int
+        $epoch : Int
             epoch to load.
-        load_optimizer_states : bool
+        $load_optimizer_states=0 : Bool
             whether to load optimizer states. Checkpoint needs
             to have been made with save_optimizer_states=True.
-        data_names : list of str
-            Default is `('data')` for a typical model used in image classification.
-        label_names : list of str
-            Default is `('softmax_label')` for a typical model used in image
+        :$data_names : array ref of str
+            Default is ['data'] for a typical model used in image classification.
+        :$label_names : array ref of str
+            Default is ['softmax_label'] for a typical model used in image
             classification.
-        logger : Logger
-            Default is `logging`.
-        context : Context or list of Context
-            Default is `cpu()`.
-        work_load_list : list of number
-            Default `None`, indicating uniform workload.
-        fixed_param_names: list of str
-            Default `None`, indicating no network parameters are fixed.
+        :$logger : Logger
+            Default is AI::MXNet::Logging.
+        :$context : Context or list of Context
+            Default is cpu(0).
+        :$work_load_list : array ref of number
+            Default is undef, indicating an uniform workload.
+        :$fixed_param_names: array ref of str
+            Default is undef, indicating no network parameters are fixed.
 =cut
 
 method load(
@@ -323,33 +301,31 @@ method load(
     %kwargs
 )
 {
-    my ($sym, $args, $auxs) = load_checkpoint($prefix, $epoch);
+    my ($sym, $args, $auxs) = __PACKAGE__->load_checkpoint($prefix, $epoch);
     my $mod = $self->new(symbol => $sym, %kwargs);
     $mod->_p->_arg_params($args);
     $mod->_p->_aux_params($auxs);
-    if($mod->params_initialized)
+    $mod->params_initialized(1);
+    if($load_optimizer_states)
     {
-        if($load_optimizer_states)
-        {
-            $mod->_p->_preload_opt_states(sprintf('%s-%04d.states', $prefix, $epoch));
-        }
+        $mod->_p->_preload_opt_states(sprintf('%s-%04d.states', $prefix, $epoch));
     }
     return $mod;
 }
 
 =head2 save_checkpoint
 
-        Save current progress to checkpoint.
-        Use mx.callback.module_checkpoint as epoch_end_callback to save during training.
+    Save current progress to a checkpoint.
+    Use mx->callback->module_checkpoint as epoch_end_callback to save during training.
 
-        Parameters
-        ----------
-        prefix : str
-            The file prefix to checkpoint to
-        epoch : int
-            The current epoch number
-        save_optimizer_states : bool
-            Whether to save optimizer states for continue training
+    Parameters
+    ----------
+    $prefix : Str
+        The file prefix to checkpoint to
+    $epoch : Int
+        The current epoch number
+    $save_optimizer_states=0 : Bool
+        Whether to save optimizer states for later training
 =cut
 
 
@@ -373,16 +349,16 @@ method save_checkpoint(Str $prefix, Int $epoch, Bool $save_optimizer_states=0)
 
     Parameters
     ----------
-    prefix : str
+    $prefix : Str
         Prefix of model name.
-    epoch : int
+    $epoch : Int
         The epoch number of the model.
-    symbol : AI::MXNet::Symbol
+    $symbol : AI::MXNet::Symbol
         The input symbol
-    arg_params : hash ref of str to AI::MXNet::NDArray
-        Model parameter, hash ref of name to AI::MXNet::NDArray of net's weights.
-    aux_params : hash ref of str to NDArray
-        Model parameter, hash ref of name to AI::MXNet::NDArray of net's auxiliary states.
+    $arg_params : HashRef[AI::MXNet::NDArray]
+        Model's parameters, hash ref of name to AI::MXNet::NDArray of net's weights.
+    $aux_params : HashRef[AI::MXNet::NDArray]
+        Model's parameters, hash ref of name to AI::MXNet::NDArray of net's auxiliary states.
     Notes
     -----
     - prefix-symbol.json will be saved for symbol.
@@ -415,74 +391,39 @@ method _reset_bind()
     $self->_p->_label_shapes(undef);
 }
 
-=head2 data_names
+method data_names()
+{
+    return $self->_p->_data_names;
+}
 
-        A list of names for data required by this module.
-=cut
-
-=head2 output_names
-
-        A list of names for data required by this module.
-=cut
+method label_names()
+{
+    return $self->_p->_label_names;
+}
 
 method output_names()
 {
     return $self->_p->_output_names;
 }
 
-=head2 data_shapes
-
-        Get data shapes.
-        Returns
-        -------
-        A list of AI::MXNet::DataDesc objects.
-=cut
-
 method data_shapes()
 {
     assert($self->binded);
     return $self->_p->_data_shapes;
 }
 
-=head2 data_shapes
-
-        Get label shapes.
-        Returns
-        -------
-        A list of AI::MXNet::DataDesc objects. The return value could be undef if
-        the module does not need labels, or if the module is not binded for
-        training (in this case, label information is not available).
-=cut
-
 method label_shapes()
 {
     assert($self->binded);
     return $self->_p->_label_shapes;
 }
 
-=head2 output_shapes
-
-        Get output shapes.
-        Returns
-        -------
-        A list of AI::MXNet::DataDesc objects.
-=cut
-
 method output_shapes()
 {
     assert($self->binded);
     return $self->_p->_exec_group->get_output_shapes;
 }
 
-=head2 get_params
-
-        Get current parameters.
-        Returns
-        -------
-        `(arg_params, aux_params)`, each a dictionary of name to parameters (in
-        `NDArray`) mapping.
-=cut
-
 method get_params()
 {
     assert($self->binded and $self->params_initialized);
@@ -493,60 +434,24 @@ method get_params()
     return ($self->_p->_arg_params, $self->_p->_aux_params);
 }
 
-=head2 init_params
-
-        Initialize the parameters and auxiliary states.
-
-        Parameters
-        ----------
-        initializer : Initializer
-            Called to initialize parameters if needed.
-        arg_params : dict
-            If not None, should be a dictionary of existing arg_params. Initialization
-            will be copied from that.
-        aux_params : dict
-            If not None, should be a dictionary of existing aux_params. Initialization
-            will be copied from that.
-        allow_missing : bool
-            If true, params could contain missing values, and the initializer will be
-            called to fill those missing params.
-        force_init : bool
-            If true, will force re-initialize even if already initialized.
-=cut
-
 method init_params(
     Maybe[AI::MXNet::Initializer]      :$initializer=AI::MXNet::Initializer->Uniform(scale => 0.01),
     Maybe[HashRef[AI::MXNet::NDArray]] :$arg_params=,
     Maybe[HashRef[AI::MXNet::NDArray]] :$aux_params=,
     Bool                               :$allow_missing=0,
-    Bool                               :$force_init=0
+    Bool                               :$force_init=0,
+    Bool                               :$allow_extra=0
 )
 {
     if($self->params_initialized and not $force_init)
     {
+        AI::MXNet::Logging->warning(
+            "Parameters already initialized and force_init=0. "
+            ."init_params call ignored."
+        );
         return;
     }
     assert($self->binded, 'call bind before initializing the parameters');
-    if(not defined $self->_p->_arg_params)
-    {
-        my @param_arrays = (
-            map { AI::MXNet::NDArray->zeros($_->[0]->shape, dtype => $_->[0]->dtype) }
-            @{ $self->_p->_exec_group->_p->param_arrays }
-        );
-        my %arg_params;
-        @arg_params{ @{ $self->_p->_param_names } } = @param_arrays;
-        $self->_p->_arg_params(\%arg_params);
-    }
-    if(not defined $self->_p->_aux_params)
-    {
-        my @aux_arrays = (
-            map { AI::MXNet::NDArray->zeros($_->[0]->shape, dtype => $_->[0]->dtype) }
-            @{ $self->_p->_exec_group->_p->aux_arrays }
-        );
-        my %aux_params;
-        @aux_params{ @{ $self->_p->_aux_names } } = @aux_arrays;
-        $self->_p->_aux_params(\%aux_params);
-    }
     my $_impl = sub {
             my ($name, $arr, $cache) = @_;
             # Internal helper for parameter initialization
@@ -569,13 +474,13 @@ method init_params(
                     }
                     if(defined $initializer)
                     {
-                        &{$initializer}($name, $arr);
+                        $initializer->($name, $arr);
                     }
                 }
             }
             else
             {
-                &{$initializer}($name, $arr) if defined $initializer;
+                $initializer->($name, $arr) if defined $initializer;
             }
     };
     my $attrs = $self->_symbol->attr_dict;
@@ -603,33 +508,64 @@ method init_params(
     $self->_p->_params_dirty(0);
 
     # copy the initialized parameters to devices
-    $self->_p->_exec_group->set_params($self->_p->_arg_params, $self->_p->_aux_params);
+    $self->_p->_exec_group->set_params($self->_p->_arg_params, $self->_p->_aux_params, $allow_extra);
+}
+
+method set_params(
+    HashRef[AI::MXNet::NDArray]  $arg_params,
+    HashRef[AI::MXNet::NDArray]  $aux_params,
+    Bool                        :$allow_missing=0,
+    Bool                        :$force_init=1,
+    Bool                        :$allow_extra=0
+)
+{
+    if(not $allow_missing)
+    {
+        $self->init_params(
+            arg_params    => $arg_params,    aux_params => $aux_params,
+            allow_missing => $allow_missing, force_init => $force_init,
+            allow_extra   => $allow_extra
+        );
+        return;
+    }
+
+    if($self->params_initialized and not $force_init)
+    {
+        AI::MXNet::Logging->warning(
+            "Parameters already initialized and force_init=False. "
+            ."set_params call ignored."
+        );
+        return;
+    }
+    $self->_p->_exec_group->set_params($arg_params, $aux_params, $allow_extra);
+    $self->_p->_params_dirty(1);
+    $self->params_initialized(1);
 }
 
 =head2 bind
 
-        Bind the symbols to construct executors. This is necessary before one
-        can perform computation with the module.
+    Bind the symbols to construct executors. This is necessary before one
+    can perform computation with the module.
 
-        Parameters
-        ----------
-        data_shapes : ArrayRef[AI::MXNet::DataDesc|NameShape]
-            Typically is `data_iter->provide_data`.
-        label_shapes : Maybe[ArrayRef[AI::MXNet::DataDesc|NameShape]]
-            Typically is `data_iter->provide_label`.
-        for_training : bool
-            Default is 1. Whether the executors should be bind for training.
-        inputs_need_grad : bool
-            Default is 0. Whether the gradients to the input data need to be computed.
-            Typically this is not needed. But this might be needed when implementing composition
-            of modules.
-        force_rebind : bool
-            Default is 0. This function does nothing if the executors are already
-            binded. But with this 1, the executors will be forced to rebind.
-        shared_module : Module
-            Default is undef. This is used in bucketing. When not undef, the shared module
-            essentially corresponds to a different bucket -- a module with different symbol
-            but with the same sets of parameters (e.g. unrolled RNNs with different lengths).
+    Parameters
+    ----------
+    :$data_shapes : ArrayRef[AI::MXNet::DataDesc|NameShape]
+        Typically is $data_iter->provide_data.
+    :$label_shapes : Maybe[ArrayRef[AI::MXNet::DataDesc|NameShape]]
+        Typically is $data_iter->provide_label.
+    :$for_training : bool
+        Default is 1. Whether the executors should be bind for training.
+    :$inputs_need_grad : bool
+        Default is 0. Whether the gradients to the input data need to be computed.
+        Typically this is not needed. But this might be needed when implementing composition
+        of modules.
+    :$force_rebind : bool
+        Default is 0. This function does nothing if the executors are already
+        binded. But with this 1, the executors will be forced to rebind.
+    :$shared_module : Module
+        Default is undef. This is used in bucketing. When not undef, the shared module
+        essentially corresponds to a different bucket -- a module with different symbol
+        but with the same sets of parameters (e.g. unrolled RNNs with different lengths).
 =cut
 
 method bind(
@@ -639,7 +575,8 @@ method bind(
     Bool                                           :$inputs_need_grad=0,
     Bool                                           :$force_rebind=0,
     Maybe[AI::MXNet::Module]                       :$shared_module=,
-    GradReq|HashRef[GradReq]|ArrayRef[GradReq]     :$grad_req='write'
+    GradReq|HashRef[GradReq]|ArrayRef[GradReq]     :$grad_req='write',
+    Maybe[ArrayRef[Str]]                           :$state_names=$self->_p->_state_names
 )
 {
     # force rebinding is typically used when one want to switch from
@@ -662,25 +599,11 @@ method bind(
     {
         assert(not $inputs_need_grad);
     }
-    $self->_p->_data_shapes([
-        map {
-            blessed $_ ? $_ : AI::MXNet::DataDesc->new(name => $_->[0], shape => $_->[1])
-        } @{ $data_shapes }
-    ]);
-
-    if($label_shapes)
-    {
-        $self->_p->_label_shapes([
-            map {
-                blessed $_ ? $_ : AI::MXNet::DataDesc->new(name => $_->[0], shape => $_->[1]) 
-            } @{ $label_shapes }
-        ]);
-    }
-    else
-    {
-        $self->_p->_label_shapes(undef);
-    }
-
+    ($data_shapes, $label_shapes) = $self->_parse_data_desc(
+        $self->data_names, $self->label_names, $data_shapes, $label_shapes
+    );
+    $self->_p->_data_shapes($data_shapes);
+    $self->_p->_label_shapes($label_shapes);
     my $shared_group;
     if($shared_module)
     {
@@ -696,6 +619,7 @@ method bind(
             data_shapes       => $self->_p->_data_shapes,
             label_shapes      => $self->_p->_label_shapes,
             param_names       => $self->_p->_param_names,
+            state_names       => $state_names,
             for_training      => $for_training,
             inputs_need_grad  => $inputs_need_grad,
             shared_group      => $shared_group,
@@ -716,6 +640,24 @@ method bind(
         # so automatically copy the already initialized params
         $self->_p->_exec_group->set_params($self->_p->_arg_params, $self->_p->_aux_params);
     }
+    else
+    {
+        assert(not defined $self->_p->_arg_params and not $self->_p->_aux_params);
+        my @param_arrays = (
+            map { AI::MXNet::NDArray->zeros($_->[0]->shape, dtype => $_->[0]->dtype, stype => $_->[0]->stype) }
+            @{ $self->_p->_exec_group->_p->param_arrays }
+        );
+        my %arg_params;
+        @arg_params{ @{ $self->_p->_param_names } } = @param_arrays;
+        $self->_p->_arg_params(\%arg_params);
+        my @aux_arrays = (
+            map { AI::MXNet::NDArray->zeros($_->[0]->shape, dtype => $_->[0]->dtype) }
+            @{ $self->_p->_exec_group->_p->aux_arrays }
+        );
+        my %aux_params;
+        @aux_params{ @{ $self->_p->_aux_names } } = @aux_arrays;
+        $self->_p->_aux_params(\%aux_params);
+    }
     if($shared_module and $shared_module->optimizer_initialized)
     {
         $self->borrow_optimizer($shared_module)
@@ -724,43 +666,29 @@ method bind(
 
 =head2 reshape
 
-        Reshape the module for new input shapes.
-        Parameters
-        ----------
-        data_shapes : ArrayRef[AI::MXNet::DataDesc]
-            Typically is $data_iter->provide_data.
-        label_shapes : Maybe[ArrayRef[AI::MXNet::DataDesc]]
-            Typically is $data_iter->provide_label.
+    Reshape the module for new input shapes.
+    Parameters
+    ----------
+    :$data_shapes : ArrayRef[AI::MXNet::DataDesc]
+        Typically is $data_iter->provide_data.
+    :$label_shapes= : Maybe[ArrayRef[AI::MXNet::DataDesc]]
+        Typically is $data_iter->provide_label.
 =cut
 
 method reshape(
-    ArrayRef[AI::MXNet::DataDesc]        $data_shapes,
-    Maybe[ArrayRef[AI::MXNet::DataDesc]] $label_shapes=
+    ArrayRef[AI::MXNet::DataDesc|NameShape]        :$data_shapes,
+    Maybe[ArrayRef[AI::MXNet::DataDesc|NameShape]] :$label_shapes=
 )
 {
     assert($self->binded);
+    ($data_shapes, $label_shapes) = $self->_parse_data_desc(
+        $self->data_names, $self->label_names, $data_shapes, $label_shapes
+    );
     $self->_p->_data_shapes($data_shapes);
     $self->_p->_label_shapes($label_shapes);
     $self->_p->_exec_group->reshape($self->_p->_data_shapes, $self->_p->_label_shapes);
 }
 
-=head2 init_optimizer
-
-        Install and initialize optimizers.
-
-        Parameters
-        ----------
-        kvstore : str or KVStore
-            Default `'local'`.
-        optimizer : str or Optimizer
-            Default `'sgd'`
-        optimizer_params : dict
-            Default { learning_rate => 0.01 }.
-        force_init : bool
-            Default `False`, indicating whether we should force re-initializing the
-            optimizer in the case an optimizer is already installed.
-=cut
-
 method init_optimizer(
     Str|AI::MXNet::KVStore :$kvstore='local',
     Optimizer              :$optimizer='sgd',
@@ -774,6 +702,10 @@ method init_optimizer(
         $self->logger->warning('optimizer already initialized, ignoring...');
         return;
     }
+    if($self->_p->_params_dirty)
+    {
+        $self->_sync_params_from_devices;
+    }
 
     my ($kvstore, $update_on_kvstore) = _create_kvstore(
         $kvstore,
@@ -832,7 +764,7 @@ method init_optimizer(
         # copy initialized local parameters to kvstore
         _initialize_kvstore(
             kvstore           => $kvstore,
-            param_arrays      => $self->_p->_exec_group->param_arrays,
+            param_arrays      => $self->_p->_exec_group->_p->param_arrays,
             arg_params        => $self->_p->_arg_params,
             param_names       => $self->_p->_param_names,
             update_on_kvstore => $update_on_kvstore
@@ -857,12 +789,12 @@ method init_optimizer(
 
 =head2 borrow_optimizer
 
-        Borrow optimizer from a shared module. Used in bucketing, where exactly the same
-        optimizer (esp. kvstore) is used.
+    Borrow optimizer from a shared module. Used in bucketing, where exactly the same
+    optimizer (esp. kvstore) is used.
 
-        Parameters
-        ----------
-        shared_module : Module
+    Parameters
+    ----------
+    shared_module : AI::MXNet::Module
 =cut
 
 method borrow_optimizer(AI::MXNet::Module $shared_module)
@@ -875,51 +807,58 @@ method borrow_optimizer(AI::MXNet::Module $shared_module)
     $self->optimizer_initialized(1);
 }
 
-=head2 forward
-
-        Forward computation.
-
-        Parameters
-        ----------
-        data_batch : DataBatch
-            Could be anything with similar API implemented.
-        is_train : bool
-            Default is `None`, which means `is_train` takes the value of `self.for_training`.
-=cut
-
 method forward(
     AI::MXNet::DataBatch $data_batch,
     Maybe[Bool]         :$is_train=
 )
 {
     assert($self->binded and $self->params_initialized);
+
+    my @curr_data_shapes = map { $_->shape } @{ $self->data_shapes };
+    my @new_data_shapes  = map { $_->shape } @{ $data_batch->data };
+    if(Data::Dumper->Dump(\@curr_data_shapes) ne Data::Dumper->Dump(\@new_data_shapes))
+    {
+        my $new_dshape;
+        if($data_batch->can('provide_data') and $data_batch->provide_data)
+        {
+            $new_dshape = $data_batch->provide_data;
+        }
+        else
+        {
+            $new_dshape = [];
+            for(zip($self->data_shapes, \@new_data_shapes)) {
+                my ($i, $shape) = @$_;
+                push @{ $new_dshape }, AI::MXNet::DataDesc->new(
+                    $i->name, $shape, $i->dtype, $i->layout
+                );
+            }
+        }
+        my $new_lshape;
+        if($data_batch->can('provide_label') and $data_batch->provide_label)
+        {
+            $new_lshape = $data_batch->provide_label;
+        }
+        elsif($data_batch->can('label') and $data_batch->label)
+        {
+            $new_lshape = [];
+            for(zip($self->label_shapes, $data_batch->label)) {
+                my ($i, $j) = @$_;
+                push @{ $new_lshape }, AI::MXNet::DataDesc->new(
+                    $i->name, $j->shape, $i->dtype, $i->layout
+                );
+            }
+        }
+        $self->reshape(data_shapes => $new_dshape, label_shapes => $new_lshape);
+    }
     $self->_p->_exec_group->forward($data_batch, $is_train);
 }
 
-=head2 backward
-
-        Backward computation.
-
-        Parameters
-        ----------
-        out_grads : NDArray or list of NDArray, optional
-            Gradient on the outputs to be propagated back.
-            This parameter is only needed when bind is called
-            on outputs that are not a loss function.
-=cut
-
 method backward(Maybe[AI::MXNet::NDArray|ArrayRef[AI::MXNet::NDArray]] $out_grads=)
 {
     assert($self->binded and $self->params_initialized);
     $self->_p->_exec_group->backward($out_grads);
 }
 
-=head2 update
-
-        Update parameters according to the installed optimizer and the gradients computed
-        in the previous forward-backward batch.
-=cut
-
 method update()
 {
     assert($self->binded and $self->params_initialized and $self->optimizer_initialized);
@@ -927,9 +866,10 @@ method update()
     if($self->_p->_update_on_kvstore)
     {
         _update_params_on_kvstore(
-            $self->_p->_exec_group->param_arrays,
-            $self->_p->_exec_group->grad_arrays,
-            $self->_p->_kvstore
+            $self->_p->_exec_group->_p->param_arrays,
+            $self->_p->_exec_group->_p->grad_arrays,
+            $self->_p->_kvstore,
+            $self->_p->_exec_group->param_names
         );
     }
     else
@@ -939,70 +879,36 @@ method update()
             $self->_p->_exec_group->_p->grad_arrays,
             $self->_p->_updater,
             scalar(@{ $self->_p->_context}),
-            $self->_p->_kvstore
+            $self->_p->_kvstore,
+            $self->_p->_exec_group->param_names
         );
     }
 }
 
-=head2 get_optputs
-
-        Get outputs of the previous forward computation.
-
-        Parameters
-        ----------
-        merge_multi_context : bool
-            Default is `True`. In the case when data-parallelism is used, the outputs
-            will be collected from multiple devices. A `True` value indicate that we
-            should merge the collected results so that they look like from a single
-            executor.
-
-        Returns
-        -------
-        If `merge_multi_context` is `True`, it is like `[out1, out2]`. Otherwise, it
-        is like `[[out1_dev1, out1_dev2], [out2_dev1, out2_dev2]]`. All the output
-        elements are `NDArray`.
-=cut
-
 method get_outputs(Bool $merge_multi_context=1)
 {
     assert($self->binded and $self->params_initialized);
     return $self->_p->_exec_group->get_outputs($merge_multi_context);
 }
 
-=head2 get_input_grads
-
-        Get the gradients with respect to the inputs of the module.
-
-        Parameters
-        ----------
-        merge_multi_context : bool
-            Default is `True`. In the case when data-parallelism is used, the outputs
-            will be collected from multiple devices. A `True` value indicate that we
-            should merge the collected results so that they look like from a single
-            executor.
-
-        Returns
-        -------
-        If `merge_multi_context` is `True`, it is like `[grad1, grad2]`. Otherwise, it
-        is like `[[grad1_dev1, grad1_dev2], [grad2_dev1, grad2_dev2]]`. All the output
-        elements are `NDArray`.
-=cut
-
 method get_input_grads(Bool $merge_multi_context=1)
 {
     assert($self->binded and $self->params_initialized and $self->inputs_need_grad);
     return $self->_p->_exec_group->get_input_grads($merge_multi_context);
 }
 
-=head2 update_metric
-        Evaluate and accumulate evaluation metric on outputs of the last forward computation.
+method get_states(Bool $merge_multi_context=1)
+{
+    assert($self->binded and $self->params_initialized);
+    return $self->_p->_exec_group->get_states($merge_multi_context);
+}
+
+method set_states(:$states=, :$value=)
+{
+    assert($self->binded and $self->params_initialized);
+    return $self->_p->_exec_group->set_states($states, $value);
+}
 
-        Parameters
-        ----------
-        eval_metric : AI::MXNet::EvalMetric
-        labels : ArrayRef[AI::MXNet::NDArray]
-            Typically $data_batch->label.
-=cut
 method update_metric(
     AI::MXNet::EvalMetric $eval_metric,
     ArrayRef[AI::MXNet::NDArray] $labels
@@ -1013,26 +919,17 @@ method update_metric(
 
 =head2 _sync_params_from_devices
 
-        Synchronize parameters from devices to CPU. This function should be called after
-        calling `update` that updates the parameters on the devices, before one can read the
-        latest parameters from `self._arg_params` and `self._aux_params`.
+    Synchronize parameters from devices to CPU. This function should be called after
+    calling 'update' that updates the parameters on the devices, before one can read the
+    latest parameters from $self->_arg_params and $self->_aux_params.
 =cut
 
 method _sync_params_from_devices()
 {
     $self->_p->_exec_group->get_params($self->_p->_arg_params, $self->_p->_aux_params);
+    $self->_p->_params_dirty(0);
 }
 
-=head2 save_optimizer_states
-
-        Save optimizer (updater) state to file
-
-        Parameters
-        ----------
-        fname : str
-            Path to output states file.
-=cut
-
 method save_optimizer_states(Str $fname)
 {
     assert($self->optimizer_initialized);
@@ -1048,16 +945,6 @@ method save_optimizer_states(Str $fname)
     }
 }
 
-=head2 load_optimizer_states
-
-        Load optimizer (updater) state from file
-
-        Parameters
-        ----------
-        fname : str
-            Path to input states file.
-=cut
-
 method load_optimizer_states(Str $fname)
 {
     assert($self->optimizer_initialized);
@@ -1075,20 +962,26 @@ method load_optimizer_states(Str $fname)
     }
 }
 
-=head2 install_monitor
-
-        Install monitor on all executors.
-
-        Paramters
-        ---------
-        AI::MXNet::Monitor
-=cut
-
 method install_monitor(AI::MXNet::Monitor $mon)
 {
     assert($self->binded);
     $self->_p->_exec_group->install_monitor($mon);
 }
 
+method _updater()
+{
+    $self->_p->_updater;
+}
 
-1;
\ No newline at end of file
+method _kvstore()
+{
+    $self->_p->_kvstore;
+}
+
+method _arg_params()
+{
+    $self->_p->_arg_params;
+}
+
+
+1;
